@@ -1,65 +1,149 @@
+import importlib
 import re
-from typing import List, Optional, Union, cast
+from types import ModuleType
+from typing import Dict, Iterable, List, Optional, Tuple, TypeVar, Union, cast
 
 from pyspark.sql import Column, DataFrame, SparkSession
 from pyspark.sql import functions as f
-from pyspark.sql.types import DataType, StructField, StructType
+
+from spark_frame.conf import PROJECT_NAME
 
 StringOrColumn = Union[str, Column]
+K = TypeVar("K")
+V = TypeVar("V")
+
+MAX_JAVA_INT = 2147483647
+
+
+class AnalysisException(Exception):
+    pass
+
+
+def is_sub_field(sub_field: str, field: str) -> bool:
+    """Return True if `sub_field` is a sub-field of `field`
+
+    >>> is_sub_field("a", "a")
+    True
+    >>> is_sub_field("a", "b")
+    False
+
+    >>> is_sub_field("a.b", "a")
+    True
+    >>> is_sub_field("a.b", "b")
+    False
+
+    >>> is_sub_field("a", "a.b")
+    False
+
+    """
+    return sub_field == field or sub_field.startswith(field + ".")
+
+
+def is_sub_field_of_any(sub_field: str, fields: List[str]) -> bool:
+    """Return True if `sub_field` is a sub-field of any field in `fields`
+
+    >>> is_sub_field_of_any("a", ["a", "b"])
+    True
+    >>> is_sub_field_of_any("a", ["b", "c"])
+    False
+
+    >>> is_sub_field_of_any("a.b", ["a", "b"])
+    True
+    >>> is_sub_field_of_any("a.b", ["b", "c"])
+    False
+
+    >>> is_sub_field_of_any("a", ["a.b"])
+    False
+
+    >>> is_sub_field_of_any("a", [])
+    False
+
+    """
+    return any([is_sub_field(sub_field, field) for field in fields])
+
+
+def group_by_key(items: Iterable[Tuple[K, V]]) -> Dict[K, List[V]]:
+    """Group the values of a list of tuples by their key.
+
+    Args:
+        items: An iterable of tuples (key, value).
+
+    Returns:
+        A dictionary where the keys are the keys from the input tuples,
+        and the values are lists of the corresponding values.
+
+    Examples:
+        >>> items = [('a', 1), ('b', 2), ('a', 3), ('c', 4), ('b', 5)]
+        >>> group_by_key(items)
+        {'a': [1, 3], 'b': [2, 5], 'c': [4]}
+        >>> group_by_key([])
+        {}
+    """
+    result: Dict[K, List[V]] = {}
+    for key, value in items:
+        if key in result:
+            result[key].append(value)
+        else:
+            result[key] = [value]
+    return result
 
 
 def quote(col: str) -> str:
     """Puts the given column name into quotes.
 
     This is useful in particular when some column names contains dots,
-    which usually happens after using the :func:`flatten` operation.
+    which usually happens after using the `flatten` transformation.
+
+    Args:
+        col: A column name
+
+    Returns:
+        The column name put inside quotes
 
     Examples:
+        >>> quote('a')
+        '`a`'
+        >>> quote('a.b')
+        '`a.b`'
 
-    >>> quote('a')
-    '`a`'
-    >>> quote('a.b')
-    '`a.b`'
+        Column names can even contain backquotes, by escaping backquotes with another backquote.
 
-    Column names can even contain backquotes, by escaping backquotes with another backquote.
+        >>> quote('a`b')
+        '`a``b`'
 
-    >>> quote('a`b')
-    '`a``b`'
-
-    Differs from `quote_idempotent` on this specific case
-    >>> quote('`a`')
-    '```a```'
-
-    :param col:
-    :return:
+        Differs from `quote_idempotent` on this specific case
+        >>> quote('`a`')
+        '```a```'
     """
     return "`" + col.replace("`", "``") + "`"
 
 
 def quote_idempotent(col: str) -> str:
-    """Puts the given column name into quotes if it is not already
+    """Puts the given column name into quotes if it is not already.
 
     This is useful in particular when some column names contains dots,
-    which usually happens after using the :func:`flatten` operation.
+    which usually happens after using the `flatten` transformation.
+
+    Args:
+        col: A column name
+
+    Returns:
+        The column name put inside quotes if it is not already
 
     Examples:
+        >>> quote_idempotent('a')
+        '`a`'
+        >>> quote_idempotent('a.b')
+        '`a.b`'
 
-    >>> quote_idempotent('a')
-    '`a`'
-    >>> quote_idempotent('a.b')
-    '`a.b`'
+        Column names can even contain backquotes, by escaping backquotes with another backquote.
 
-    Column names can even contain backquotes, by escaping backquotes with another backquote.
+        >>> quote_idempotent('a`b')
+        '`a``b`'
 
-    >>> quote_idempotent('a`b')
-    '`a``b`'
-
-    Differs from `quote` on this specific case
-    >>> quote_idempotent('`a`')
-    '`a`'
-
-    :param col:
-    :return:
+        Differs from `quote` on this specific case
+        >>> quote_idempotent('`a`')
+        '`a`'
     """
     return quote(unquote(col))
 
@@ -70,8 +154,12 @@ def quote_columns(columns: List[str]) -> List[str]:
     This is useful in particular when some column names contains dots,
     which usually happens after using the :func:`flatten` operation.
 
-    :param columns:
-    :return:
+    Args:
+        columns: A list of column names
+
+    Returns:
+        A list with each column name put into quotes
+
     """
     return [quote(col) for col in columns]
 
@@ -79,22 +167,24 @@ def quote_columns(columns: List[str]) -> List[str]:
 def unquote(col_name: str) -> str:
     """Removes quotes from a quoted column name.
 
+    Args:
+        col_name: A column name
+
+    Returns:
+        The column name with quotes removed
+
     Examples:
+        >>> unquote('a')
+        'a'
+        >>> unquote('`a`')
+        'a'
+        >>> unquote('`a.b`')
+        'a.b'
 
-    >>> unquote('a')
-    'a'
-    >>> unquote('`a`')
-    'a'
-    >>> unquote('`a.b`')
-    'a.b'
+        Column names can even contain backquotes, by escaping backquotes with another backquote.
 
-    Columns names can even contain backquotes, by escaping backquotes with another backquote.
-
-    >>> unquote('`a``b`')
-    'a`b'
-
-    :param col_name:
-    :return:
+        >>> unquote('`a``b`')
+        'a`b'
     """
     if col_name[0] == "`" and col_name[-1] == "`":
         col_name = col_name[1:-1]
@@ -121,46 +211,54 @@ __split_col_name_regex = re.compile(
 def split_col_name(col_name: str) -> List[str]:
     """Splits a Spark column name representing a nested field into a list of path parts.
 
+    Args:
+        col_name: A column name
+
+    Returns:
+        The column name split by nesting level
+
     Examples:
-    In this example: `a` is a struct containing a field `b`.
+        In this example: `a` is a struct containing a field `b`.
 
-    >>> split_col_name("a.b")
-    ['a', 'b']
-    >>> split_col_name("ab")
-    ['ab']
+        >>> split_col_name("a.b")
+        ['a', 'b']
+        >>> split_col_name("ab")
+        ['ab']
 
-    Field names can contain dots when escaped between backquotes (`)
+        Field names can contain dots when escaped between backquotes (`)
 
-    >>> split_col_name("`a.b`")
-    ['a.b']
-    >>> split_col_name("`a.b`.`c.d`")
-    ['a.b', 'c.d']
-    >>> split_col_name("`a.b`.c.`d`")
-    ['a.b', 'c', 'd']
+        >>> split_col_name("`a.b`")
+        ['a.b']
+        >>> split_col_name("`a.b`.`c.d`")
+        ['a.b', 'c.d']
+        >>> split_col_name("`a.b`.c.`d`")
+        ['a.b', 'c', 'd']
 
-    Field names can even contain backquotes, by escaping backquotes with another backquote.
+        Field names can even contain backquotes, by escaping backquotes with another backquote.
 
-    >>> split_col_name("`a``b`")
-    ['a`b']
-    >>> split_col_name("`.a.``.b.`")
-    ['.a.`.b.']
-    >>> split_col_name("`ab`.`c``d`.fg")
-    ['ab', 'c`d', 'fg']
-
-    :param col_name:
-    :return:
+        >>> split_col_name("`a``b`")
+        ['a`b']
+        >>> split_col_name("`.a.``.b.`")
+        ['.a.`.b.']
+        >>> split_col_name("`ab`.`c``d`.fg")
+        ['ab', 'c`d', 'fg']
     """
     col_parts = [unquote(match[0]) for match in __split_col_name_regex.findall(col_name)]
     return col_parts
 
 
-def str_to_col(args: StringOrColumn) -> Column:
+def str_to_col(col: StringOrColumn) -> Column:
     """Converts string or Column argument to Column types
 
     Requires the SparkSession to be instantiated.
 
-    Examples:
+    Args:
+        col: A column name or Column expression
 
+    Returns:
+        A Column expression
+
+    Examples:
         >>> from pyspark.sql import SparkSession
         >>> spark = SparkSession.builder.appName("doctest").getOrCreate()
         >>> str_to_col("id")
@@ -169,15 +267,14 @@ def str_to_col(args: StringOrColumn) -> Column:
         Column<'COUNT(1)'>
         >>> str_to_col("*")
         Column<'unresolvedstar()'>
-
     """
-    if isinstance(args, str):
-        return f.col(args)
+    if isinstance(col, str):
+        return f.col(col)
     else:
-        return args
+        return col
 
 
-def strip_margin(text: str):
+def strip_margin(text: str) -> str:
     """For every line in this string, strip a leading prefix consisting of whitespace, tabs and carriage returns
     followed by | from the line.
 
@@ -191,7 +288,6 @@ def strip_margin(text: str):
         A stripped string
 
     Examples:
-
         >>> print(strip_margin('''
         ...     |a
         ...     |b
@@ -215,47 +311,6 @@ def strip_margin(text: str):
         return s
 
 
-def get_nested_col_type_from_schema(col_name: str, schema: StructType) -> DataType:
-    """Fetch recursively the DataType of a column inside a DataFrame schema (or more generally any StructType)
-
-    Example:
-    >>> from pyspark.sql import SparkSession
-    >>> spark = SparkSession.builder.appName("doctest").getOrCreate()
-    >>> df = spark.createDataFrame([
-    ...      (1, {"a.b" : {"c.d": 1, "e": "1", "f`g": True}}),
-    ...      (2, {"a.b" : {"c.d": 2, "e": "2", "f`g": True}}),
-    ...      (3, {"a.b" : {"c.d": 3, "e": "3", "f`g": True}}),
-    ...   ],
-    ...   "id INT, `a.b` STRUCT<`c.d`:INT, e:STRING, `f``g`:BOOLEAN>"
-    ... )
-    >>> get_nested_col_type_from_schema("`a.b`.`c.d`",df.schema).simpleString()
-    'int'
-    >>> get_nested_col_type_from_schema("`a.b`.e",df.schema).simpleString()
-    'string'
-    >>> get_nested_col_type_from_schema("`a.b`.`f``g`",df.schema).simpleString()
-    'boolean'
-
-    :param schema: the DataFrame schema (or StructField) in which the column type will be fetched.
-    :param col_name: the name of the column to get
-    :return:
-    """
-    col_parts = split_col_name(col_name)
-
-    def get_col(col: str, fields: List[StructField]) -> StructField:
-        for field in fields:
-            if field.name == col:
-                return field
-        raise ValueError(f'Cannot resolve column name "{col_name}"')
-
-    struct: Union[StructType, DataType] = schema
-    for col_part in col_parts:
-        assert_true(isinstance(struct, StructType))
-        struct = cast(StructType, struct)
-        struct = get_col(col_part, struct.fields).dataType
-    assert_true(isinstance(struct, DataType))
-    return cast(DataType, struct)
-
-
 def get_instantiated_spark_session() -> SparkSession:
     """Get the instantiated SparkSession. Raises an AssertionError if it does not exists."""
     optional_spark = getattr(SparkSession, "_instantiatedSession")
@@ -266,7 +321,7 @@ def get_instantiated_spark_session() -> SparkSession:
 
 def show_string(df: DataFrame, n: int = 20, truncate: Union[bool, int] = True, vertical: bool = False) -> str:
     """Write the first ``n`` rows to the console into a string.
-    This is similar to [DataFrame.show](pyspark.sql.DataFrame.show) except it returns a string instead of directly
+    This is similar to [DataFrame.show][pyspark.sql.DataFrame.show] except it returns a string instead of directly
     writing it to stdout.
 
     Args:
@@ -281,7 +336,6 @@ def show_string(df: DataFrame, n: int = 20, truncate: Union[bool, int] = True, v
         A string representing the first `n` rows of the DataFrame
 
     Examples:
-
         >>> from pyspark.sql import SparkSession
         >>> spark = SparkSession.builder.appName("doctest").getOrCreate()
         >>> df = spark.sql('''
@@ -331,7 +385,7 @@ def show_string(df: DataFrame, n: int = 20, truncate: Union[bool, int] = True, v
 
 def schema_string(df: DataFrame) -> str:
     """Write the DataFrame schema to a string.
-    This is similar to [DataFrame.printSchema](pyspark.sql.DataFrame.printSchema) except it returns a string instead
+    This is similar to [DataFrame.printSchema][pyspark.sql.DataFrame.printSchema] except it returns a string instead
     of directly writing it to stdout.
 
     Args:
@@ -358,16 +412,7 @@ def schema_string(df: DataFrame) -> str:
     return df._jdf.schema().treeString()
 
 
-def safe_struct_get(s: Optional[Column], field: str) -> Column:
-    """Get a column's subfield, unless the column is None, in which case it returns
-    a Column expression for the field itself."""
-    if s is None:
-        return f.col(field)
-    else:
-        return s[field]
-
-
-def assert_true(assertion: bool, error: Union[str, BaseException] = None) -> None:
+def assert_true(assertion: bool, error: Optional[Union[str, BaseException]] = None) -> None:
     """Raise an Exception with the given error_message if the assertion passed is false.
 
     !!! Tip
@@ -379,19 +424,20 @@ def assert_true(assertion: bool, error: Union[str, BaseException] = None) -> Non
         assertion: The boolean result of an assertion
         error: An Exception or a message string (in which case an AssertError with this message will be raised)
 
-    >>> assert_true(3==3, "3 <> 4")
-    >>> assert_true(3==4, "3 <> 4")
-    Traceback (most recent call last):
-    ...
-    AssertionError: 3 <> 4
-    >>> assert_true(3==4, ValueError("3 <> 4"))
-    Traceback (most recent call last):
-    ...
-    ValueError: 3 <> 4
-    >>> assert_true(3==4)
-    Traceback (most recent call last):
-    ...
-    AssertionError
+    Examples:
+        >>> assert_true(3==3, "3 <> 4")
+        >>> assert_true(3==4, "3 <> 4")
+        Traceback (most recent call last):
+        ...
+        AssertionError: 3 <> 4
+        >>> assert_true(3==4, ValueError("3 <> 4"))
+        Traceback (most recent call last):
+        ...
+        ValueError: 3 <> 4
+        >>> assert_true(3==4)
+        Traceback (most recent call last):
+        ...
+        AssertionError
     """
     if not assertion:
         if isinstance(error, BaseException):
@@ -400,3 +446,41 @@ def assert_true(assertion: bool, error: Union[str, BaseException] = None) -> Non
             raise AssertionError(error)
         else:
             raise AssertionError()
+
+
+def load_external_module(module_name: str) -> ModuleType:
+    """Load and return a Python module, raising an exception if it is not installed or does not meet the
+    expected version requirements.
+
+    Args:
+        module_name: The name of the module to load.
+
+    Returns:
+        module: The loaded Python module.
+
+    Raises:
+        ImportError: If the module is not found, does not have a `__version__` attribute, or does not meet
+            the expected version requirements.
+
+    Examples:
+        >>> load_external_module('platform').__name__
+        'platform'
+
+        >>> load_external_module('nonexistent_module')
+        Traceback (most recent call last):
+            ...
+        ImportError: Module 'nonexistent_module' not found.
+        To keep spark-frame as light, flexible and secure as possible, it was not included in its dependencies.
+        Please add it to your project dependencies to use this method.
+    """
+    try:
+        module = importlib.import_module(module_name)
+    except ImportError:
+        raise ImportError(
+            f"Module '{module_name}' not found.\n"
+            f"To keep {PROJECT_NAME} as light, flexible and secure as possible, "
+            "it was not included in its dependencies.\n"
+            "Please add it to your project dependencies to use this method."
+        )
+
+    return module
